import java.awt.Color;
import java.awt.event.KeyEvent;

import com.threed.jpct.Camera;
import com.threed.jpct.Config;
import com.threed.jpct.FrameBuffer;
import com.threed.jpct.IRenderer;
import com.threed.jpct.Lights;
import com.threed.jpct.Loader;
import com.threed.jpct.Matrix;
import com.threed.jpct.Object3D;
import com.threed.jpct.OcTree;
import com.threed.jpct.Projector;
import com.threed.jpct.SimpleVector;
import com.threed.jpct.Texture;
import com.threed.jpct.TextureInfo;
import com.threed.jpct.TextureManager;
import com.threed.jpct.World;
import com.threed.jpct.util.KeyMapper;
import com.threed.jpct.util.KeyState;
import com.threed.jpct.util.Light;
import com.threed.jpct.util.ShadowHelper;

class CarTest {

	/**
	 * Some jPCT related stuff
	 */
	private FrameBuffer buffer = null;
	private World theWorld = null;
	private TextureManager texMan = null;
	private Camera camera = null;

	/**
	 * Some things for the "game logic"
	 */
	private Object3D terrain = null;
	private Car car = null;
	private ProjectileManager bulMan = null;
	private PlantManager plantMan = null;
	private SkidMarkManager skidMan = null;

	/**
	 * The texture used for blitting the framerate
	 */
	private Texture numbers = null;
	private Texture spotLight = null;
	private Projector projector = new Projector();

	private int fps;
	private int lastFps;
	private long totalFps;

	private boolean isIdle = false;
	private boolean exit = false;

	/**
	 * Flags for the keys
	 */
	private boolean left = false;
	private boolean right = false;
	private boolean forward = false;
	private boolean back = false;
	private boolean fire = false;
	private int fireCount = 3;

	private float speed = 0;

	private KeyMapper keyMapper = null;

	private ShadowHelper shadows = null;

	/**
	 * Very complex stuff...impossible to explain...
	 */
	public static void main(String[] args) {
		new CarTest(args);
	}

	/**
	 * The constructor. Here we are initializing things...
	 */
	private CarTest(String[] args) {

		Config.maxPolysVisible = 10000;

		isIdle = false;
		totalFps = 0;
		fps = 0;
		lastFps = 0;

		/**
		 * Initialize the World instance and get the TextureManager (a
		 * singleton)
		 */
		theWorld = new World();
		texMan = TextureManager.getInstance();

		Config.farPlane = 800;

		theWorld.getLights().setOverbrightLighting(Lights.OVERBRIGHT_LIGHTING_DISABLED);
		theWorld.getLights().setRGBScale(Lights.RGB_SCALE_2X);
		theWorld.setAmbientLight(35, 35, 35);

		/**
		 * Load the textures needed and add them to the TextureManager. We are
		 * loading the "numbers" texture for blitting the framerate and the
		 * texture for the ground. Textures used by the game entities are loaded
		 * by them.
		 */

		numbers = new Texture("textures/other/numbers.jpg");
		texMan.addTexture("numbers", numbers);

		spotLight = new Texture("textures/light/spot.jpg", true);
		texMan.addTexture("spotlight", spotLight);
		spotLight.setProjector(projector, true);
		projector.rotateX((float) Math.PI / 2f);
		projector.setFOVLimits(0, 2);
		projector.setFOV(2);

		Texture rocks = new Texture("textures/rocks.jpg");
		texMan.addTexture("rocks", rocks);

		TextureInfo ti = new TextureInfo(texMan.getTextureID("rocks"));
		ti.add(texMan.getTextureID("spotlight"), TextureInfo.MODE_MODULATE);

		/**
		 * Load the map (i.e. the terrain or "ground")
		 */
		Object3D[] objs = Loader.load3DS("models/terascene.3ds", 400);
		if (objs.length > 0) {
			terrain = objs[0];
			terrain.setTexture(ti);
		}

		terrain.enableLazyTransformations();
		theWorld.addObject(terrain);

		/**
		 * The terrain isn't located where we want it to, so we take care of
		 * this here:
		 */
		SimpleVector pos = terrain.getCenter();
		pos.scalarMul(-1f);
		terrain.translate(pos);
		terrain.rotateX((float) -Math.PI / 2f);
		terrain.translateMesh();
		terrain.rotateMesh();
		terrain.setTranslationMatrix(new Matrix());
		terrain.setRotationMatrix(new Matrix());

		/**
		 * Setup the octree and the collision mode/listener for the terrain
		 */
		OcTree oc = new OcTree(terrain, 500, OcTree.MODE_OPTIMIZED);
		terrain.setOcTree(oc);
		oc.setCollisionUse(OcTree.COLLISION_USE);
		oc.setRenderingUse(OcTree.RENDERING_DONT_USE);

		/**
		 * The game logic relies on "Managers" in this example. We are
		 * instantiating them here.
		 */
		bulMan = new ProjectileManager(theWorld);
		skidMan = new SkidMarkManager(theWorld);

		Config.collideOffset = 250;

		terrain.setCollisionMode(Object3D.COLLISION_CHECK_OTHERS);
		terrain.setCollisionOptimization(Object3D.COLLISION_DETECTION_OPTIMIZED);
		terrain.addCollisionListener(new BulletTerrainListener(bulMan));

		Light light = new Light(theWorld);
		light.setIntensity(150, 150, 150);
		light.setAttenuation(-1);
		light.setPosition(new SimpleVector(0, -500, 0));

		/**
		 * Setup some optimizations for outdoor rendering
		 */
		Config.tuneForOutdoor();

		/**
		 * Do some AWT setup work
		 */
		initializeFrame();

		Projector proj = new Projector();
		proj.setPosition(new SimpleVector(0, -350, 0));
		proj.rotateX((float) Math.PI / 2f);
		shadows = new ShadowHelper(theWorld, buffer, projector, 2048);

		shadows.addReceiver(terrain);
		shadows.setAmbientLight(new Color(40, 40, 40));
		plantMan = new PlantManager(theWorld, terrain, 1800, shadows);

		/**
		 * We want to drive around...a car may help
		 */
		car = new Car(shadows);
		car.addToWorld(theWorld);

		/**
		 * The game entities are building themselves, so we only have to build
		 * the terrain here
		 */
		terrain.build();
		terrain.compile();

		/**
		 * Place the camera at the starting position.
		 */
		camera = theWorld.getCamera();
		camera.setPosition(0, -2500, -1500);
		camera.lookAt(car.getTransformedCenter());

		/**
		 * Here we go...!
		 */
		gameLoop();
	}

	/**
	 * This initializes the AWT frame either in fullscreen or in windowed mode.
	 * This is not a waterproof intialization, but i didn't want to do a AWT
	 * tutorial here (and i would be the wrong person to do this anyway...:-)).
	 * Change whatever you think that needs change here...
	 */
	private void initializeFrame() {
		buffer = new FrameBuffer(800, 600, FrameBuffer.SAMPLINGMODE_HARDWARE_ONLY);
		buffer.enableRenderer(IRenderer.RENDERER_OPENGL, IRenderer.MODE_OPENGL);
		buffer.disableRenderer(IRenderer.RENDERER_SOFTWARE);
		keyMapper = new KeyMapper();
	}

	/**
	 * The fps counter is blitted into the rendered framebuffer here and the
	 * rendering results will be displayed (hence the name of the method...:-))
	 */
	private void display() {
		blitNumber((int) totalFps, 5, 2);
		plantMan.drawRadar(buffer, car);
		buffer.displayGLOnly();
	}

	/**
	 * A simple method that does the number-blitting.
	 */
	private void blitNumber(int number, int x, int y) {
		if (numbers != null) {
			String sNum = Integer.toString(number);
			for (int i = 0; i < sNum.length(); i++) {
				char cNum = sNum.charAt(i);
				int iNum = cNum - 48;
				buffer.blit(numbers, iNum * 5, 0, x, y, 5, 9, FrameBuffer.TRANSPARENT_BLITTING);
				x += 5;
			}
		}
	}

	/**
	 * Does the collision detection and the movement of the car.
	 */
	private void moveCar() {

		if (left) {
			car.turnLeft();
		}
		if (right) {
			car.turnRight();
		}

		/**
		 * We need to store this matrix for the case that the car can't be
		 * placed at the current location.
		 */
		Matrix backUpTrans = car.getTranslationMatrix().cloneMatrix();

		if (forward) {
			if (speed < 15) {
				speed += 0.1f;
			}
			car.setSpeed(speed);
			car.moveForward();
		}
		if (back) {
			if (speed > 0) {
				speed -= 0.3f;
			}
			if (speed < 0) {
				speed = 0;
			}
			car.setSpeed(speed);
			car.moveForward();
		}

		if (speed >= 0 && !back && !forward) {
			/**
			 * If car is still moving but not accelerating anymore, then it
			 * should get slower over time.
			 */
			speed -= 0.075f;
			if (speed < 0) {
				speed = 0;
			}
			car.setSpeed(speed);
			car.moveForward();
		}

		/**
		 * Try to place the car at its new location
		 */
		boolean ok = car.place(terrain);
		if (!ok) {
			/**
			 * Placing the car failed...we'll restore the backup translation
			 * matrix in that case (may happen when actually leaving the map).
			 */
			car.setTranslationMatrix(backUpTrans);
			speed = 0;
			car.setSpeed(speed);
		} else {
			/**
			 * The car has been placed, so we can draw some skidmarks if
			 * required.
			 */
			skidMan.createSkidMarks(car);
		}

		SimpleVector pos = car.getTransformedCenter();
		pos.y -= 200;

		projector.setPosition(pos);
	}

	/**
	 * Move the camera. The camera will always look at the car from behind and
	 * takes the last camera position into account, so that the movement gets a
	 * smooth feeling.
	 */
	private void moveCamera() {
		SimpleVector oldCamPos = camera.getPosition();
		SimpleVector oldestCamPos = new SimpleVector(oldCamPos);
		oldCamPos.scalarMul(9f);

		SimpleVector carCenter = car.getTransformedCenter();
		SimpleVector camPos = new SimpleVector(carCenter);
		SimpleVector zOffset = car.getZAxis();
		SimpleVector yOffset = new SimpleVector(0, -200, 0);
		zOffset.scalarMul(-350f);

		camPos.add(zOffset);
		camPos.add(yOffset);

		camPos.add(oldCamPos);
		camPos.scalarMul(0.1f);

		SimpleVector delta = camPos.calcSub(oldestCamPos);
		float len = delta.length();

		if (len != 0) {
			/**
			 * Do a collision detection between the camera and the ground to
			 * prevent the camera from moving into the ground.
			 */
			theWorld.checkCameraCollisionEllipsoid(delta.normalize(), new SimpleVector(20, 20, 20), len, 3);
		}

		/**
		 * And finally: Look at the car
		 */
		camera.lookAt(car.getTransformedCenter());
	}

	/**
	 * Generates and/or move existing bullets
	 */
	private void processProjectiles() {
		if (fire || fireCount != 3) {
			fireCount--;
			if (fireCount == 0) {
				bulMan.createBullet(car);
				fireCount = 3;
			}
		}
		bulMan.moveBullets();
	}

	/**
	 * This is the game's main loop. We are doing some additional setup work
	 * here and rendering the scene in a loop, as well as handling the movement
	 * and the collision detection before each frame.
	 */
	private void gameLoop() {
		World.setDefaultThread(Thread.currentThread());

		Timer timer = new Timer(25);
		timer.start();

		Timer fpsTimer = new Timer(1000);
		fpsTimer.start();

		while (!exit) {
			if (!isIdle) {

				long ticks = timer.getElapsedTicks();

				for (int i = 0; i < ticks; i++) {
					moveCar();
					processProjectiles();
					moveCamera();
				}

				poll();

				projector.lookAt(car.getTransformedCenter());

				shadows.updateShadowMap();

				buffer.clear();
				shadows.drawScene();
				buffer.update();
				display();

				fps++;

				if (fpsTimer.getElapsedTicks() > 0) {
					totalFps = (fps - lastFps);
					lastFps = fps;
				}

				Thread.yield();

			} else {
				try {
					Thread.sleep(500);
				} catch (InterruptedException e) {
				}
			}
		}
		System.exit(0);
	}

	/**
	 * Use the KeyMapper to poll the keyboard
	 */
	private void poll() {
		KeyState state = null;
		do {
			state = keyMapper.poll();
			if (state != KeyState.NONE) {
				keyAffected(state);
			}
		} while (state != KeyState.NONE);
	}

	private void keyAffected(KeyState state) {
		int code = state.getKeyCode();
		boolean event = state.getState();

		switch (code) {
		case (KeyEvent.VK_ESCAPE):
			exit = event;
			break;
		case (KeyEvent.VK_LEFT):
			left = event;
			break;
		case (KeyEvent.VK_RIGHT):
			right = event;
			break;
		case (KeyEvent.VK_UP):
			forward = event;
			break;
		case (KeyEvent.VK_SPACE):
			fire = event;
			break;
		case (KeyEvent.VK_DOWN):
			back = event;
			break;
		case (KeyEvent.VK_L):
			if (event) {
				spotLight.setEnabled(!spotLight.isEnabled());
			}
			break;
		}
	}

	private class Timer {

		private long ticks = 0;
		private long granularity = 0;

		public Timer(int granularity) {
			this.granularity = granularity;
		}

		public void start() {
			ticks = System.currentTimeMillis();
		}

		public long getElapsedTicks() {
			long cur = System.currentTimeMillis();
			long l = cur - ticks;

			if (l >= granularity) {
				ticks = cur - (l % granularity);
				return l / granularity;
			}
			return 0;
		}
	}

}
